/*
 * $Id: atacmds.c 705412 2015-03-31 11:59:28Z ib-builder $
 *
 * Copyright (C) 1999-2000 Michael Cornwell <cornwell@acm.org>
 * Copyright (C) 2000 Andre Hedrick <andre@linux-ide.org>
 *
 * Copyright (c) 2002-2012, Juniper Networks, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * You should have received a copy of the GNU General Public License
 * (for example COPYING); if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/jnx/dkio.h>
#include <sys/ata.h>
#include <sys/fcntl.h>
#include <sys/syslog.h>
#include <sys/time.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>

#include "atacmds.h"
#include "smartd.h"
#include "extern.h"

/* Buffer used for ioctl */
#define SMARTD_ATA_BUF_SZ   512
#define SMARTD_ATA_TIMEOUT  3600    /* 28 - 60 minutes needed for smarttest */ 
u_char buf[SMARTD_ATA_BUF_SZ];

static int 
try_ioctl (int device, unsigned long cmd, struct dk_iocmd* params) 
{
    unsigned int no_of_tries = 0;
    do {
        if (ioctl (device, cmd, params) != -1) return 0;
	/*
	 * Note, please do not modify the errno variable after the ioctl call 
	 * above, else functions calling try_ioctl may not work correctly
	 */
        if (errno != EAGAIN) {
                return errno;
        }
        sleep(2);
    } while (++no_of_tries < 200); 
    return EAGAIN;
}

int 
atareadhdidentity (int device, struct ata_params *atap)
{
    if (ioctl (device , IOCATAGPARM, atap) != 0 ) { 
	logging(LOG_ERR, "%s: ioctl: %s", __FUNCTION__, strerror(errno));
	return -1; 
    }
    return 0;
}

int 
ataversioninfo (struct ata_params *drive)
{
    int i, power;
    int atavalue = 0; 

    if ((drive->version_major == 0xffff) || (drive->version_major == 0x000)) {
    	return -1;
    }
    for (i = 1, power = 2 ; i < 16; power <<= 1, i++) {
	if (drive->version_major &  power) {
	    atavalue = i;
	}
    }
    return atavalue;
}

int 
atasmartsupport (struct ata_params *drive)
{
    if (drive->support.command1 & ATA_SUPPORT_SMART) {

        if (ataversioninfo (drive) > 5) {
	    if ((*((u_int16_t*)(&drive->support)) == 0xffff) ||
		 (*((u_int16_t*)(&drive->support)) == 0x0000)) {
		return (-1); /* word 82 not implemented on drive per ATA spec */
	    }

	    if (drive->enabled.command1 & ATA_SUPPORT_SMART) {
		return (2); /* drive supports SMART and is enabled */
	    } else {
		return (1); /* drive supports SMART and is disabled */
	    }
	} else {
	    return (255); /* drive supports SMART and stat is unknown*/
	}
    } else {
	return (-1); /* drive does not support SMART */
    }
}

int 
atapwrmgmtsupport (struct ata_params *drive)
{
    if (drive->support.command1 & ATA_SUPPORT_POWERMGT) {

        if (ataversioninfo (drive) > 5) {
	    if ((*((u_int16_t*)(&drive->support)) == 0xffff) ||
		 (*((u_int16_t*)(&drive->support)) == 0x0000)) {
		return (-1); /* word 82 not implemented on drive per ATA spec */
	    }

	    if (drive->enabled.command1 & ATA_SUPPORT_POWERMGT) {
		return (2); /* drive supports advance power mgmt and is enabled */
	    } else {
		return (1); /* drive supports advance power mgmt and is disabled */
	    }
	} else {
	    return (255); /* drive supports advance power mgmt and stat is unknown*/
	}
    } else {
	return (-1); /* drive does not support advance power mgmt */
    }
}

static void
dumpbuf(void) {
    int i,j;
    unsigned char pbuf[100];
    unsigned char *p;

    i=0;
    while (i < 512) {
        p=pbuf;
        sprintf(p, "%3d: ", i);
        p+=5;
        for (j=0; j<16; j++) {
            sprintf(p,"%2x ", buf[i]);
            p+=3;
            i++;
        }
        *p = 0;
	logging(LOG_ERR, "%s", pbuf);
    };
}

/*
 * Issue a SMART control command
 */
static int 
ata_smart_cntrl_cmd (int device, u_int8_t sub_cmd)
{    
    struct ata_ioc_request request;
    
    bzero(buf, sizeof(buf));
    request.u.ata.command = WIN_SMART;
    request.timeout = SMARTD_ATA_TIMEOUT;

    request.u.ata.feature = sub_cmd;
    request.u.ata.lba = (0xc24f<<8); 
    request.count = 0; 
    request.flags = ATA_CMD_CONTROL;

    if ((ioctl (device , IOCATAREQUEST,  &request) != 0) || request.error) {
	logging(LOG_ERR, "%s: ioctl: %s, request.error %d", __FUNCTION__, strerror(errno),
		request.error);
	return -1;
    }
    return 0;
}

int 
atareadsmartvalues (int device, struct ata_smart_values *data)
{    
    int i;
    u_char chksum;
    struct ata_ioc_request request;

    
    bzero(buf, sizeof(buf));
    request.u.ata.command = WIN_SMART;
    request.timeout = SMARTD_ATA_TIMEOUT;

    request.u.ata.feature = SMART_READ_VALUES;
    request.u.ata.lba = 0xc24f<<8; /* This Magic number was in ata-disk.c */
    request.flags = ATA_CMD_READ;
    request.data = buf;
    request.count = sizeof(buf);

    if ((ioctl (device , IOCATAREQUEST,  &request) != 0) || request.error) {
	/* REVISIT : Need a way to check for command abort */
        /* error register  has ABRT set, command not supported */
        if (errno == EIO && (buf[ERROR_PLACEHOLDER] & ATA_ABORT)) {
            return ENODEV;
        }
	logging(LOG_ERR, "%s: ioctl: %s", __FUNCTION__, strerror(errno));
	return -1;
    }
    
    /* Unsigned add for first 511 bytes, ingnoring the overflow */
    for (i = chksum = 0; i < 511; i++)
	chksum+= buf[i];

    /* Take 2's complement */
    chksum = ~chksum;
    chksum += 1;

    if (chksum != buf[511]) {
	logging(LOG_ERR, "%s:Smart Read Failed, Chksum(%2x) error,"
			    " expected(%2x)", __FUNCTION__, buf[511], chksum);
	dumpbuf();
	return -1;
    }

    memcpy(data, buf, sizeof(*data));
    return 0;
}

int 
atareadselftestlog (int device, struct ata_smart_selftestlog *data)
{    
    struct ata_ioc_request request;

    
    bzero(buf, sizeof(buf));
    request.u.ata.command = WIN_SMART;
    request.timeout = SMARTD_ATA_TIMEOUT;

    request.u.ata.feature = SMART_READ_LOG_SECTOR;
    request.u.ata.lba = 6 | 0xc24f<<8; 
    request.u.ata.count = 1;
    request.flags = ATA_CMD_READ;
    request.data = buf;
    request.count = sizeof(buf);

    if ((ioctl (device , IOCATAREQUEST,  &request) != 0) || request.error) {
	logging(LOG_ERR, "%s: ioctl: %s, request.error %d", __FUNCTION__, strerror(errno),
		request.error);
	return -1;
    }

    memcpy(data, buf, sizeof(*data));
    return 0;
}

int 
atareaderrorlog (int device, struct ata_smart_errorlog *data)
{    
    struct ata_ioc_request request;

    bzero(buf, sizeof(buf));
    request.u.ata.command = WIN_SMART;
    request.timeout = SMARTD_ATA_TIMEOUT;

    request.u.ata.feature = SMART_READ_LOG_SECTOR;
    request.u.ata.lba = 0x01 | 0xc24f<<8; 
    request.u.ata.count = 1;
    request.flags = ATA_CMD_READ;
    request.data = buf;
    request.count = sizeof(buf);

    if ((ioctl (device , IOCATAREQUEST,  &request) != 0) || request.error) {
	logging(LOG_ERR, "%s: ioctl: %s, request.error %d", __FUNCTION__, strerror(errno),
		request.error);
	return -1;
    }

    memcpy(data, buf, sizeof(*data));
    return 0;
}

int 
atareadsmartthresholds (int device, struct ata_smart_thresholds *data)
{    
    struct ata_ioc_request request;
    
    bzero(buf, sizeof(buf));
    request.u.ata.command = WIN_SMART;
    request.timeout = SMARTD_ATA_TIMEOUT;

    request.u.ata.feature = SMART_READ_THRESHOLDS;
    request.u.ata.lba = 1 | (0xc24f<<8); 
    request.u.ata.count = 1;
    request.flags = ATA_CMD_READ;
    request.data = buf;
    request.count = sizeof(buf);

    if ((ioctl (device , IOCATAREQUEST,  &request) != 0) || request.error) {
	logging(LOG_ERR, "%s: ioctl: %s, request.error %d", __FUNCTION__, strerror(errno),
		request.error);
	return -1;
    }

    memcpy(data, &buf, sizeof(*data));
    return 0;
}

int 
atasetsmartthresholds (int device, struct ata_smart_thresholds *data)
{    
    struct ata_ioc_request request;
    
    bzero(buf, sizeof(buf));
    request.u.ata.command = WIN_SMART;
    request.timeout = SMARTD_ATA_TIMEOUT;

    request.u.ata.feature = 0xD7;
    request.u.ata.lba = 1 | (0xc24f<<8); 
    request.u.ata.count = 1;
    request.flags = ATA_CMD_WRITE;
    request.data = buf;
    request.count = sizeof(buf);

    if ((ioctl (device , IOCATAREQUEST,  &request) != 0) || request.error) {
	logging(LOG_ERR, "%s: ioctl: %s, request.error %d", __FUNCTION__, strerror(errno),
		request.error);
	return -1;
    }

    memcpy(data, &buf, sizeof(*data));
    return 0;
}


int 
ataenablesmart (int device)
{    
    return ata_smart_cntrl_cmd (device, SMART_ENABLE);
}

int 
atadisablesmart (int device)
{  
    return ata_smart_cntrl_cmd (device, SMART_DISABLE);
}

int 
ataenableautooffline (int device)
{    
    struct ata_ioc_request request;
    
    bzero(buf, sizeof(buf));
    request.u.ata.command = WIN_SMART;
    request.timeout = SMARTD_ATA_TIMEOUT;

    request.u.ata.feature = SMART_AUTO_OFFLINE;
    request.u.ata.lba = 248 | (0xc24f<<8); 
    request.flags = ATA_CMD_CONTROL;

    if ((ioctl (device , IOCATAREQUEST,  &request) != 0) || request.error) {
	logging(LOG_ERR, "%s: ioctl: %s, request.error %d", __FUNCTION__, strerror(errno),
		request.error);
	return -1;
    }
    return 0;
}

int 
atadisableautooffline (int device)
{    
    return ata_smart_cntrl_cmd (device, SMART_AUTO_OFFLINE );
}

int 
atasmartstatus (int device)
{    
    struct ata_ioc_request request;
    unsigned const char normal_lo=0x4f, normal_hi=0xc2;
    unsigned const char failed_lo=0xf4, failed_hi=0x2c;
    unsigned char low,high;
    
    bzero(buf, sizeof(buf));
    request.u.ata.command = WIN_SMART;
    request.timeout = SMARTD_ATA_TIMEOUT;
    request.u.ata.feature = SMART_STATUS;
    request.u.ata.lba = 1 | (0xc24f<<8);  /* REVISIT define for 0xc24f */
    request.flags = ATA_CMD_CONTROL;
    request.count = 0;

    if ((ioctl (device , IOCATAREQUEST,  &request) != 0) || request.error) {
	logging(LOG_ERR, "%s: ioctl: %s, request.error %d", __FUNCTION__, strerror(errno),
		request.error);
	return -1;
    }
    high = (request.u.ata.lba >> 16) & 0xff;
    low = (request.u.ata.lba >> 8) & 0xff;
    
    /* Cyl low and Cyl high unchanged means "Good SMART status" */
    if (low==normal_lo && high==normal_hi)
      return 0;
    
    /* These values mean "Bad SMART status" */
    if (low==failed_lo && high==failed_hi)
      return 1;
    
    return 1;
}

int 
atasmarttest (int device, int testtype)
{  
    struct ata_ioc_request request;
    
    bzero(buf, sizeof(buf));
    request.u.ata.command = WIN_SMART;
    request.timeout = SMARTD_ATA_TIMEOUT;

    request.u.ata.feature = SMART_IMMEDIATE_OFFLINE;
    request.u.ata.lba = testtype | (0xc24f<<8); 
    request.flags = ATA_CMD_CONTROL;
    request.count = 0;

    if ((ioctl (device , IOCATAREQUEST,  &request) != 0) || request.error) {
	logging(LOG_ERR, "%s: ioctl: %s, request.error %d", __FUNCTION__, strerror(errno),
		request.error);
	return -1;
    }
    return 0;
}

int 
atacheckdevicestatus (int device, int *error)
{    
    struct dk_iocmd parms = {0, 0, 0 , 0 , 0, 0, 0, buf };
    
    *error = 0;

    if (try_ioctl (device , DKIO_GET_STATUS, &parms) != 0) {
	if (errno == EBUSY)
	    return EBUSY;
	else {
	    logging(LOG_ERR, "%s: ioctl: %s", __FUNCTION__, strerror(errno));
	    return -1;
	}
    }

    *error = buf[ERROR_PLACEHOLDER];
    return 0;
}

int 
ataformatunit (int device, struct ata_params *drive)
{    
    struct ata_ioc_request request;
    
    bzero(buf, sizeof(buf));
    request.u.ata.command = SEC_ERASE_PREPARE ;
    request.timeout = 3600;  /* REVISIT: This may not enough */

    request.u.ata.feature = 0;
    request.u.ata.lba = 0; 
    request.flags = ATA_CMD_CONTROL;
  
    /* word 83 not implemented on drive per ATA spec */
    if ((*(((u_int16_t*)(&drive->support))+1) == 0xffff) || (*(((u_int16_t*)(&drive->support))+1) == 0x0000)) {
	return EOPNOTSUPP;
    }

    /* Security feature set not implemented */
    if (!(drive->support.command1 & ATA_SUPPORT_SECURITY)) {
	return EOPNOTSUPP;
    }

    if ((ioctl (device , IOCATAREQUEST,  &request) != 0) || request.error) {
	logging(LOG_ERR, "Unable to issue security erase prepare command"
		" (error reg = %2x)", buf[ERROR_PLACEHOLDER]);
	return -1;
    }

    request.u.ata.command = FORMAT_UNIT;
    request.u.ata.feature = 0x11;

    if ((ioctl (device , IOCATAREQUEST,  &request) != 0) || request.error) {
	/*
	 * REVISIT check error code 
	 * Need to check fro command abort
	 * if (errno != EIO || !(buf[ERROR_PLACEHOLDER] & ATA_ABORT)) {
	 *   logging(LOG_ERR, "Unable to issue format command"
         *		    " (error reg = %2x)", buf[ERROR_PLACEHOLDER]);
	 *   return -1;
	 * else
	 *   	
	 *   return EOPNOTSUPP;
	 * }
	 */
    }

    setbuf(stdout, NULL);
    printf("Formating command completed successfuly.\n");
    return 0;
}

int 
atasmartofflinetest (int device)
{    
    return atasmarttest(device, OFFLINE_FULL_SCAN );
}

int 
atasmartshortselftest (int device)
{    
    return atasmarttest(device, SHORT_SELF_TEST );
}

int 
atasmartextendselftest (int device)
{    
    return atasmarttest(device, EXTEND_SELF_TEST );
}

int 
atasmartshortcapselftest (int device)
{    
    return atasmarttest(device, SHORT_CAPTIVE_SELF_TEST);
}

int 
atasmartextendcapselftest (int device)
{
    return atasmarttest(device, EXTEND_CAPTIVE_SELF_TEST);
}

int 
atasmartselftestabort (int device)
{
    return atasmarttest(device, 127);
}

/* Test Time Functions */

int 
isofflinetesttime (struct ata_smart_values data)
{
    return (int) data.total_time_to_complete_off_line;
}

int 
isshortselftesttime (struct ata_smart_values data)
{
    return (int) data.short_test_completion_time;
}

int 
isextendedselftesttime (struct ata_smart_values data)
{
    return (int) data.extend_test_completion_time;
}

int 
issmarterrorlogcapable (struct ata_smart_values data)
{
    return data.errorlog_capability & 0x01;
}

int 
issupportexecuteofflineimmediate (struct ata_smart_values data)
{
    return data.offline_data_collection_capability & 0x01;
}

int 
issupportautomatictimer (struct ata_smart_values data)
{
    return data.offline_data_collection_capability & 0x02;
}

int 
issupportofflineabort (struct ata_smart_values data)
{
    return data.offline_data_collection_capability & 0x04;
}

int 
issupportofflinesurfacescan (struct ata_smart_values data)
{
    return data.offline_data_collection_capability & 0x08;
}

int 
issupportselftest (struct ata_smart_values data)
{
    return data.offline_data_collection_capability & 0x10;
}

int 
atachecksmart (struct ata_smart_values data, struct ata_smart_thresholds thresholds)
{
    int i;

    for (i = 0 ; i < NUMBER_ATA_SMART_ATTRIBUTES ; i++) {
	if ((data.vendor_attributes[i].id !=0) &&   
	     (thresholds.thres_entries[i].id != 0) &&
	     (data.vendor_attributes[i].status_flag & 0x1) &&
	     (data.vendor_attributes[i].normalized <
	       thresholds.thres_entries[i].normalized_threshold) &&
	     (thresholds.thres_entries[i].normalized_threshold != 0xFE)) {
	     return data.vendor_attributes[i].id;
	}
    }
#ifdef DEBUG_FAILURE
    return 200;
#else
    return 0;
#endif
}
int 
ataissuestandby (int device, unsigned int timeinterval)
{    
    struct ata_ioc_request request;

    bzero(buf, sizeof(buf));
    request.u.ata.command = ATA_STANDBY_CMD;
    request.timeout = SMARTD_ATA_TIMEOUT;

    request.u.ata.feature = 0;
    request.u.ata.lba = 0; 
    request.u.ata.count = timeinterval/5; 
    request.flags = ATA_CMD_CONTROL;
    request.count = 0; 

    if ((ioctl (device , IOCATAREQUEST,  &request) != 0) || request.error) {
	logging(LOG_ERR, "%s: ioctl: %s, request.error %d", __FUNCTION__, strerror(errno),
		request.error);
	return -1;
    }
    return 0;
}

int 
ataissuestandbyimmediate (int device)
{    
    
    struct ata_ioc_request request;

    
    bzero(buf, sizeof(buf));
    request.u.ata.command = ATA_STANDBY_IMMEDIATE;
    request.timeout = SMARTD_ATA_TIMEOUT;

    request.u.ata.feature = 0;
    request.u.ata.lba = 0; 
    request.flags = ATA_CMD_CONTROL;
    request.data = buf;
    request.count = 0;

    if ((ioctl (device , IOCATAREQUEST,  &request) != 0) || request.error) {
	logging(LOG_ERR, "%s: ioctl: %s, request.error %d", __FUNCTION__, strerror(errno),
		request.error);
	return -1;
    }
    return 0;
}


/*
 * Check disk power mode. In case of error we assume disk is in active
 * or idle state and not in standby
 */
char
atacheckpowermode (int device, unsigned char powercheckenabled)
{
    char powermode = ATA_POWERMODE_ACTIVE_OR_IDLE;
    struct ata_ioc_request request;
    
    /* 
     * If power-check is not enabled, query the status from the 
     * kernel
     */
    if (!powercheckenabled) {
	struct ata_adp_sync_req sync_req;

	sync_req.cmd = ATAADPPOWERMODE;
	if (ioctl ( device , IOCATAADPSYNC,  &sync_req) != 0) {
	    logging(LOG_ERR, "%s: ioctl: %s", __FUNCTION__, strerror(errno));
	} else {
	    powermode = (sync_req.arm_params.in_standby) ?
			ATA_POWERMODE_STANDBY :
			ATA_POWERMODE_ACTIVE_OR_IDLE;
	}
	return powermode;
    }

    /* Power check is enabled, so query the status from the device */
    bzero(buf, sizeof(buf));
    request.u.ata.command = ATA_CHECK_POWERMODE;
    request.timeout = SMARTD_ATA_TIMEOUT;

    request.u.ata.feature = 0;
    request.u.ata.lba = 0; 
    request.flags = ATA_CMD_CONTROL;
    request.data = buf;
    request.count = 0;

    if ((ioctl (device , IOCATAREQUEST,  &request) != 0) || request.error) {
	logging(LOG_ERR, "%s: ioctl: %s, request.error %d", __FUNCTION__, strerror(errno),
		request.error);
        return powermode;
    }
    powermode = request.u.ata.count;
    return powermode;
}

int
atastandbyarmset(int device, char arm_standby)
{
    struct  ata_adp_sync_req sync_req;

    if (arm_standby) 
	sync_req.cmd = ATAADPSYNCARM;
    else 
	sync_req.cmd = ATAADPSYNCUNARM;
    
    if (ioctl (device , IOCATAADPSYNC,  &sync_req) != 0) {
        logging(LOG_ERR, "%s: ioctl: %s", __FUNCTION__, strerror(errno));
        return -1;
    }

    return 0;
}

char
atastandbyarmget(int device, ad_arm_t *arm_info)
{
    struct  ata_adp_sync_req sync_req;

    sync_req.cmd = ATAADPSYNCGET;
    if (ioctl (device , IOCATAADPSYNC,  &sync_req) != 0) {
        logging(LOG_ERR, "%s: ioctl: %s", __FUNCTION__, strerror(errno));
        return -1;
    }

    if (arm_info) {
	memcpy((void *)arm_info, &(sync_req.arm_params), sizeof(*arm_info));
    }

    return arm_info->arm_standby;
}

/*
 * Enable/Disable write caching to disk
 */
int
atawritecacheset (int device, char enable)
{    
    struct ata_ioc_request request;
    
    bzero(buf, sizeof(buf));
    request.u.ata.command = ATA_SETFEATURES;
    request.timeout = SMARTD_ATA_TIMEOUT;

    request.u.ata.lba = 0; 
    request.flags = ATA_CMD_CONTROL;
    request.data = buf;
    request.count = 0;

    if (enable) {
	request.u.ata.feature = ATA_SF_ENAB_WCACHE;
    } else {
	request.u.ata.feature = ATA_SF_DIS_WCACHE;
    }

    if ((ioctl (device , IOCATAREQUEST,  &request) != 0) || request.error) {
	logging(LOG_ERR, "%s: ioctl: %s, request.error %d", __FUNCTION__, strerror(errno),
		request.error);
        return -1;
    }

    return 0;
}

int 
ataissueidle (int device, unsigned int timeinterval)
{
    struct ata_ioc_request request;
    
    bzero(buf, sizeof(buf));
    request.u.ata.command = ATA_IDLE_CMD;
    request.timeout = SMARTD_ATA_TIMEOUT;

    request.u.ata.lba = 0; 
    request.flags = ATA_CMD_CONTROL;
    request.data = buf;
    request.u.ata.count = timeinterval/5;
    request.u.ata.feature = 0;
    request.count = 0;

    if ((ioctl (device , IOCATAREQUEST,  &request) != 0) || request.error) {
	logging(LOG_ERR, "%s: ioctl: %s, request.error %d", __FUNCTION__, strerror(errno),
		request.error);
        return -1;
    }
    return 0;
}

static int
version(int version)
{       
    int bit;

    if (version == 0xffff) {
	return 0;
    }

    for (bit = 15; bit >= 0; bit--) {
	if (version & (1 << bit)) {
	    return bit;
	}
    }

    return 0;
}

void
cap_print (struct ata_params *parm)
{
    printf("ATA/ATAPI revision    %d\n",
	    version(parm->version_major));
    printf("device model          %.40s\n", parm->model);
    printf("serial number         %.20s\n", parm->serial);
    printf("firmware revision     %.8s\n", parm->revision);

    printf("cylinders             %d\n", parm->cylinders);
    printf("heads                 %d\n", parm->heads);
    printf("sectors/track         %d\n", parm->sectors);

    printf("lba%ssupported         ", (parm->capabilities1 & ATA_SUPPORT_LBA) ?
	    " " : " not ");
    if (parm->lba_size_1) {
	printf("%d sectors\n", parm->lba_size_1);
    } else {
	printf("\n");
    }

    printf("lba48%ssupported         ", 
	   (parm->support.command2 & ATA_SUPPORT_ADDRESS48)?
	    " " : " not ");
    if (parm->lba_size48_1) {
	printf("%lld sectors\n", parm->lba_size48_1);
    } else {
	printf("\n");
    }

    printf("dma%ssupported\n", (parm->capabilities1 & ATA_SUPPORT_DMA)?
	    " " : " not");
    printf("overlap%ssupported\n", (parm->capabilities1 & ATA_SUPPORT_OVERLAP) ?
	    " " : " not ");
    printf("\nFeature                      Support  Enable"
	    "Value   Vendor\n");
    printf("write cache                    %s       %s\n",
	    (parm->support.command1 & ATA_SUPPORT_WRITECACHE) ? "yes" : "no",
	    (parm->enabled.command1 & ATA_SUPPORT_WRITECACHE) ? "yes" : "no");

    printf("read ahead                     %s       %s\n",
	    (parm->support.command1 & ATA_SUPPORT_LOOKAHEAD) ? "yes" : "no",
	    (parm->enabled.command1 & ATA_SUPPORT_LOOKAHEAD) ? "yes" : "no");

    printf("dma queued                     %s       %s      "
	    "%d/%02X\n",
	    (parm->support.command2 & ATA_SUPPORT_QUEUED) ? "yes" : "no",
	    (parm->enabled.command2 & ATA_SUPPORT_QUEUED) ? "yes" : "no",
	    ATA_QUEUE_LEN (parm->queue), ATA_QUEUE_LEN (parm->queue));

    printf("SMART                          %s       %s\n",
	    (parm->support.command1 & ATA_SUPPORT_SMART) ? "yes" : "no",                                                            
	    (parm->enabled.command1 & ATA_SUPPORT_SMART) ? "yes" : "no");                                                                

    printf("microcode download             %s       %s\n",
	    (parm->support.command2 & ATA_SUPPORT_MICROCODE) ? "yes" : "no",
	    (parm->enabled.command2 & ATA_SUPPORT_MICROCODE) ? "yes" : "no");

    printf("security                       %s       %s\n",
	    (parm->support.command1 & ATA_SUPPORT_SECURITY) ? "yes" : "no",
	    (parm->enabled.command1 & ATA_SUPPORT_SECURITY) ? "yes" : "no");

    printf("power management               %s       %s\n",
	    (parm->support.command1 & ATA_SUPPORT_POWERMGT) ? "yes" : "no",
	    (parm->enabled.command1 & ATA_SUPPORT_POWERMGT) ? "yes" : "no");

    printf("advanced power management      %s       %s      "
	    "%d/%02X\n",
	    (parm->support.command2 & ATA_SUPPORT_APM) ? "yes" : "no",
	    (parm->enabled.command2 & ATA_SUPPORT_APM) ? "yes" : "no",
	    parm->apm_value, parm->apm_value);

    printf("automatic acoustic management  %s       %s      "
	    "%d/%02X %d/%02X\n",
	    (parm->support.command2 & ATA_SUPPORT_AUTOACOUSTIC) ? "yes" : "no",
	    (parm->enabled.command2 & ATA_SUPPORT_AUTOACOUSTIC) ? "yes" : "no",
	    ATA_ACOUSTIC_CURRENT(parm->acoustic),
	    ATA_ACOUSTIC_CURRENT(parm->acoustic),
	    ATA_ACOUSTIC_VENDOR(parm->acoustic), 
	    ATA_ACOUSTIC_VENDOR(parm->acoustic));
}
