/*
 * $Id: smartd.c 810948 2016-10-20 09:44:06Z ib-builder $
 *
 * Copyright (C) 2000 Michael Cornwell <cornwell@acm.org>
 *
 * Copyright (c) 2002-2012, Juniper Networks, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * You should have received a copy of the GNU General Public License
 * (for example COPYING); if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */



#include <stdio.h>
#include <stdarg.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <sys/jnx/dkio.h>
#include <sys/ata.h>
#include <sys/sysctl.h>
#include <jnx/jnx_sysctl.h>
#include <jnx/jnx_syslog.h>
#include <jnx/filecopy.h>
#include <jnx/inttypes.h>

#include "atacmds.h"
#include "smartd.h"
#include "extern.h"

#ifndef evInitID
#define	evInitID(id) ((id)->opaque = NULL)
#endif
#ifndef evTestID
#define evTestID(id) ((id).opaque != NULL)
#endif

#define SMARTD_TRACE_FILE       "/var/log/smartd.trace"
#define TIME_BUF_SZ             50

#define DEFAULT_SSD_ID1         248
#define DEFAULT_SSD_ID2         249

#define sby_dlog(x...)                  \
    if (standbyloglevel) {              \
        logging(standbyloglevel, x);    \
    }                                   

#define sby_is_adaptive(s)      (((s)->sby_type == SBY_ADAPTIVE_ALWAYS) || \
                                 ((s)->sby_type == SBY_ADAPTIVE_IF_ENABLED))

#define sby_is_periodic(s)      (((s)->sby_type == SBY_PERIODIC_ALWAYS) || \
                                 ((s)->sby_type == SBY_PERIODIC_IF_ENABLED))
/* Static Protos */
static int driveissupported (struct ata_params *);
static void atadevicescan (atadevices_t *, char *);
static void atacomparesmartvalues (atadevices_t *, struct ata_smart_values);
static int atacheckdevice (atadevices_t *);
static void checkdevices ( atadevices_t *);
static int driveissupported (struct ata_params *);
static __inline__ void sby_trace_spin (ad_arm_t *armi);
static void sby_trace(const char *fmt, ...) __printflike(1, 2);
const char * reason2str(char reason);
static int deviceneedsstandby (atadevices_t *drive);
void readstandbyconfig (void);
static int sby_is_enabled(sby_table_t *sby);
static void smart_check (evContext ctx, void *uap,
                         struct timespec due __unused,
                         struct timespec inter __unused);
static void power_check (evContext ctx, void *uap,
                         struct timespec due __unused,
                         struct timespec inter __unused);
static void standby_request (evContext ctx, void *uap,
                             struct timespec due __unused,
                             struct timespec inter __unused);
static void daily_check (evContext ctx, void *uap,
                         struct timespec due __unused,
             struct timespec inter __unused);
static void armtimers (atadevices_t *atadevices);
/* JUNOS begin */
void readconfigfile(void);
static int parseconfigline(atadevices_t *dev, char *line, int lineno);
static int getvalue(char *arg, int min, int max, int lineno, char option);
static int parsetoken(atadevices_t *dev, char *token, int lineno);
static int findmissingopt(atadevices_t *dev, unsigned char id, int lineno);
static parse_flag_name_t * getname(unsigned char arg);
/* JUNOS end */
/* Globals */
static uint disk_inactivity_timer   = SMARTD_DISK_INACT_TMR;
static time_t standbyinterval	    = SMARTD_STANDBY_INTERVAL_DEFAULT;
static int  cli_standbyenable = FALSE;  /* TRUE if interval is enabled in CLI */
uint standbyloglevel		    = 0;
static time_t standbysummary	    = SMARTD_DAILY_CHECK;
static int  tracefd		    = -1;
static char tracebuf[BUFSIZ];
static int checksinamonth	    = 0;
static uint periodic_power_check    = 0;
static int spindown_time_valid = 0;
static long spindown_time_sec  = 0; 
static long spindown_time_msec = 0;
static struct timeval spindown_time_tv;

/* Number of ata device to scan */
int numatadevices;

/* Table of qualified models */
const char *modeltab[] = {
    "ST310212A",			/* CVP5000 */
    "ST36421A",				/* CVP5000 */
    "QUANTUM FIREBALL SE6.4A",		/* CVP5000 */
    "IBM-DARA-212000",			/* TEKNOR -RE2*/
    "IBM-DCXA-210000",			/* TEKNOR -RE2*/
    "IBM-DJSA-210",			/* TEKNOR -RE2*/
    "IBM-DJSA-220",			/* TEKNOR -RE2*/
    "TOSHIBA MK2016GAP",		/* TEKNOR -RE2*/
    "FUJITSU MHN2300AT",		/* FORCE  -RE3*/
    "FUJITSU MHL2300AT",		/* FORCE  -RE3*/
    "FUJITSU MHM2100AT",		/* FORCE  -RE3*/
    "FUJITSU MHR2020AT",		/* TEKNOR -RE2*/
    "FUJITSU MHR2030AT",		/* FORCE  -RE3*/
    "FUJITSU MHS2030AT",		/* FORCE  -RE3*/
    "FUJITSU MHS2020AT",                /* KONTRON -RE5, RE-850 */
    "FUJITSU MHS2020AT E",              /* KONTRON -RE5, RE-850 */
    "FUJITSU MHT2030AT",		/* FORCE  - RE3, RE4 */
    "FUJITSU MHT2030AR",                /* FORCE  - RE3, RE4, 24x7 */
    "HTS548020M9AT00",			/* KONTRON -RE5, RE-850 */
    "UGB94BPH32H0S1-KCI",               /* Unigen */
    "UGB94ARF32H0S3-KC",                /* Unigen */
    "SG9STL1C64GB02EMR",                /* SMART MODULAR */
    "SGC13S064-MS9RBC-EM",              /* Soligen */ 
    NULL
};


/*
 * atasbytab[]
 *
 * Use this array to modify setting for any disk that does not
 * work with default settings. Also if a new property that needs to
 * become disk model specific is needed then add it to the data
 * structure for this array.
 *
 * E.g. of disk specific properties.
 *
 * a) Hitachi disk HTS548020M9AT00 needs write cache disabled for good to
 *    prevent a lock up problem
 * b) Some disk may require that standby be turned off. In which case
 *    set SBY_NOTNEEDED as they standby type.
 * c) Some disks may require more than 5 minutes of standby time.
 * d) Some disks may need standby only if a standby interval is explicitly
 *    specified in CLI
 * e) Some disks may need standby always unless explicitly disabled in
 *    CLI
 * etc etc etc
 *
 */

/* 
 * DISABLE_WCACHE option is ignored. For Hitachi drives
 * it has already been disabled in the kernel and is not 
 * required for other drives
 *
 * Inactivity timer is this array is ignored. global
 * var disk_inactivity_timer determines value 
 */
sby_table_t    atasbytab[] = {
    /* The great Hitachi disk. Write cache should be disabled for good */
    /* No special support for hitachi */
#if 0    
    { "HTS548020M9AT00", NON_24x7, SBY_ADAPTIVE_ALWAYS, DISABLE_WCACHE,
        300000, 300, 60, NOSTOP_AT_LIMIT, (AD_INACTIVITY_TIME_PARAM * 5)},
#endif
    { "SMART CF", IS_24x7, SBY_NOTNEEDED, DISABLE_WCACHE,
        300000, 300, 60, NOSTOP_AT_LIMIT, (AD_INACTIVITY_TIME_PARAM * 5)},
    { "SAMSUNG MCBQE32G8MPP-0VA", IS_24x7, SBY_NOTNEEDED, DISABLE_WCACHE,
        300000, 300, 60, NOSTOP_AT_LIMIT, (AD_INACTIVITY_TIME_PARAM * 5)},
    { "STEC MACH-8 SSD", IS_24x7, SBY_NOTNEEDED, DISABLE_WCACHE,
        300000, 300, 60, NOSTOP_AT_LIMIT, (AD_INACTIVITY_TIME_PARAM * 5)},
    { NULL,              NON_24x7, SBY_ADAPTIVE_ALWAYS, DISABLE_WCACHE,
        300000, 300, 60, NOSTOP_AT_LIMIT, (AD_INACTIVITY_TIME_PARAM * 5)}
};

atadevices_t atadevices[MAXATADEVICES];
/* for storing the parse info */
parseinfo_t parse_info; /* JUNOS */
/*
 * Check if smartd has been qualified on the drive model.
 */
static int
driveissupported (struct ata_params * drive)
{
    int i = 0;

    while (modeltab[i] != NULL) {
	if (!strcmp(drive->model, modeltab[i])) {
	    return TRUE;
	}
	i++;
    }

    return FALSE;
}

/* JUNOS begin */
static parse_flag_name_t parse_flag_name[] = {
    {FOUND_DEVICE_NAME,      "DEVICENAME",      '\0'},
    {FOUND_NUM_OF_IDS,       "NUM_OF_IDS",       'n'},
    {FOUND_ATTRIBUTE_IDS,    "ATTRIBUTE_IDS",    'i'},
    {FOUND_THRESHOLD_LIMITS, "THRESHOLD_LIMITS", 't'},
    {FOUND_VALUE,            "VALUE",            'v'},
    {FOUND_FLAG,             "FLAG",             'f'} ,
    {FOUND_END,              "NULL",            '\0'}
};

/* Helper function for converting enum to str */
static parse_flag_name_t *
getname(unsigned char arg)
{
    parse_flag_name_t *flag_name = parse_flag_name;

    while(flag_name->name) {
	if(arg == flag_name->flag)
	    return flag_name;
	flag_name++;
    }

    return NULL;
}

/* The function finds the missing option for each ID configured.
 * Prints the name the option which was missing and for each ID.
 * INPUT value 0 is for general parse and other values are for
 * each configured ID
 * return value : 1 no missing option for any ID.
 * return value : -1 missing option for ID.
 */
static int
findmissingopt(atadevices_t *dev, unsigned char id, int lineno)
{
    unsigned char parseerror = FIND_MISSING_OPTIONS(id);
    parse_flag_name_t *flag_name;
    unsigned char missing_opt = 1;
    int ret_val = 1;

    while (parseerror) {
	if ((parseerror & missing_opt) &&
	    (flag_name = getname(missing_opt))) {
	    switch (id) {
		case 0:
		    logging(LOG_ERR,"File smartd.conf, lineno %d."
			    "%s:%s option missing."
			    "Specific %s option using -%c.\n",lineno,
			    dev->drive.model,  flag_name->name,
			    flag_name->name, flag_name->option);
		    break;
		default:
		    logging(LOG_ERR,"File smartd.conf, lineno %d."
			    "%s:%s option missing for attribute id %d."
			    "Specific %s option using -%c.\n", lineno,
			    dev->drive.model, flag_name->name, CONFIG_ID(id-1),
			    flag_name->name, flag_name->option);
		    break;
	    }
	    ret_val = -1;
	}

	parseerror &= (~missing_opt);
	missing_opt <<= 1;
    }

    return ret_val;
}


/* The function converts the arg to int and checks the range
 * return val :
 * -1 if the arg is null or value is not in the range
 * value if the arg is valid one
 */
static int
getvalue(char *arg, int min, int max, int lineno, char option)
{
    int val;
    if (!arg) {
	logging(LOG_ERR, "File smartd.conf line %d ,"
		"-%c directive not correct ",
		lineno, option);
	return -1;
    }

    if (*arg == '-') {
	logging(LOG_ERR, "File smartd.conf line %d ,"
		"missing args for -%c option",
		lineno, option);
	return -1;
    }

    val = strtol(arg, NULL, 0);
    if (!(min <= val && val <= max)) {
	logging(LOG_ERR, "File smartd.conf line %d ,"
		"-%c directive should be in the range (%d-%d)",
		lineno, option, min, max);
	return -1;
    }

    return val;
}

/* The function parses the token and updated the atrribute ids,
 * their threshold limits, values and flags.
 * return value :
 * -1 : parse error for the token
 *  1 : token parse successful
 *  0 : The SSD device does not have attribute thresholds
 */
static int
parsetoken(atadevices_t *dev, char *token, int lineno)
{
    const char *delim = " \n\t";
    char sym;
    char *arg;
    int val, num_of_attr, i;

    /* if the token is not recognized */
    if (*token!='-' || strlen(token)!=2) {
	logging(LOG_ERR, "smartd.conf line %d, unknown directive", lineno);
	return -1;
    }

    sym = token[1];
    switch (sym) {
	case 'n':
	case 'N':
	    /* the number of attributes to be monitored */
	    if (!(GET_PARSED_STATUS(FOUND_DEVICE_NAME))) {
		logging(LOG_ERR, "smartd.conf line %d, specify "
			"device series name at the start before"
			"specifing args", lineno);
		return -1;
	    }

	    if ((val = getvalue(strtok(NULL, delim), 1, 30,
				lineno, sym)) < 0) {
		return -1;
	    }

	    NO_OF_CONFIG_IDS = val; // the number of attributes
	    SET_PARSED_STATUS(FOUND_NUM_OF_IDS);
	    break;
	case 'i':
	case 'I':
	    /* the attribute id's to be monitored */
	    if (!(GET_PARSED_STATUS(FOUND_NUM_OF_IDS)) &&
		!(GET_PARSED_STATUS(FOUND_DEVICE_NAME))) {
		logging(LOG_ERR, "smartd.conf line %d, number "
			"of attributes missing specify number of attributes"
			"using -n option", lineno);
		return -1;
	    }	
	    if (IDS_PARSED >= NUMBER_ATA_SMART_ATTRIBUTES) {
		logging(LOG_ERR,"smartd.conf line %d, number "
			"of ids configured have exceeded the max(%d) "
			"attributes \n",lineno, NUMBER_ATA_SMART_ATTRIBUTES);
		return -1;
	    }

	    num_of_attr = NO_OF_CONFIG_IDS;
	    do {
		if ((val = getvalue(strtok(NULL, delim), 1, 255,
				    lineno, sym)) < 0) {
		    return -1;
		}
		for (i = 0; i < NUMBER_ATA_SMART_ATTRIBUTES; i++) {
		    if (dev->ssdthres.thres_entries[i].id == val) {
			CONFIG_ID(IDS_PARSED) = val;
			SET_ID_STATUS(IDS_PARSED, PARSED_ID);
			IDS_PARSED++;
			break;
		    }
		}
		
		if (i >= NUMBER_ATA_SMART_ATTRIBUTES) {
		    if (val == DEFAULT_SSD_ID1 || val == DEFAULT_SSD_ID2) {
			logging(LOG_NOTICE,"Device %s is either not a SSD "
				"or SSD DOES NOT HAVE THRESHOLDS",
				dev->drive.model);
			return 0;
		    }
		    logging(LOG_ERR, "smartd.conf line %d, attribute id :"
			    "%d(0x%02x)entered is wrong", lineno, val, val);
		    return -1;
		}
		num_of_attr--;
	    } while (num_of_attr > 0);

	    SET_PARSED_STATUS(FOUND_NUM_OF_IDS);
	    SET_PARSED_STATUS(FOUND_ATTRIBUTE_IDS);
	    break;
	case 't':
	case 'T':
	    /* the threshold limits for the ids to be monitored */
	    if (!(GET_PARSED_STATUS(FOUND_ATTRIBUTE_IDS))) {
		logging(LOG_ERR, "smartd.conf line %d, attributes ids "
			"missing specify attributes ids using -i "
			"option", lineno);
		return -1;
	    }
	    if (THRES_PARSED >= IDS_PARSED) {
		logging(LOG_ERR,"smartd.conf line %d, threshold limits "
			"are more than configured IDS \n", lineno);
		return -1;
	    }

	    num_of_attr = NO_OF_CONFIG_IDS;
	    do {
		if ((val = getvalue(strtok(NULL, delim), 1, 0xfffff,
				    lineno, sym)) < 0) {
		    return -1;
		}
		for (i = 0; i < NUMBER_ATA_SMART_ATTRIBUTES; i++) {
		    if (dev->ssdthres.thres_entries[i].id ==
			CONFIG_ID(THRES_PARSED)) {
			if (dev->ssdthres.thres_entries[i].threshold) {
			    val = val << 16;
			}
			dev->ssdthres.thres_entries[i].threshold |= val;
			SET_ID_STATUS(THRES_PARSED, FOUND_THRESHOLD_LIMITS);
			THRES_PARSED++;
			break;
		    }
		}
		num_of_attr--;
	    } while (num_of_attr > 0);

	    SET_PARSED_STATUS(FOUND_THRESHOLD_LIMITS);
	    break;
	case 'v':
	case 'V':
	    /* the threshold limits for the ids to be monitored
	     * against raw or normalized value
	     */
	    if (!(GET_PARSED_STATUS(FOUND_ATTRIBUTE_IDS))) {
		logging(LOG_ERR, "smartd.conf line %d, attributes ids "
			"missing specify attributes ids using -i "
			"option", lineno);
		return -1;
	    }
	    if (VAL_PARSED >= IDS_PARSED) {
		logging(LOG_ERR,"smartd.conf line %d, Value args "
			"are more than configured IDS \n", lineno);
		return -1;
	    }

	    num_of_attr = NO_OF_CONFIG_IDS;
	    do {
		arg = strtok(NULL, delim);
		if(!arg) {
		    logging(LOG_ERR, "smartd.conf line %d, parse error "
			    "missing args for -v option", lineno);
		    return -1;
		}

		if (strcasecmp("RAW", arg) == 0) {
		    val = FLAG_RAW;
		} else if (strcasecmp("NORM", arg) == 0) {
		    val = FLAG_NORM;
		} else {
		    if (*arg == '-') {
			logging(LOG_ERR, "smartd.conf line %d, parse error "
				"missing args for -v option", lineno);
		    } else {
			logging(LOG_ERR, "smartd.conf line %d, parse error "
				"value should be raw or norm for -v option",
				lineno);
		    }
		    return -1;
		}

		for (i = 0; i < NUMBER_ATA_SMART_ATTRIBUTES; i++) {
		    if (dev->ssdthres.thres_entries[i].id ==
			CONFIG_ID(VAL_PARSED)) {
			dev->ssdthres.thres_entries[i].flag |= val;
			SET_ID_STATUS(VAL_PARSED, FOUND_VALUE);
			VAL_PARSED++;
			break;
		    }
		}
		num_of_attr--;
	    } while (num_of_attr > 0);

	    SET_PARSED_STATUS(FOUND_VALUE);
	    break;
	case 'f':
	case 'F':
	    /* the threshold limits for the ids to be monitored
	     * against raw or normalized value
	     */
	    if (!(GET_PARSED_STATUS(FOUND_ATTRIBUTE_IDS))) {
		logging(LOG_ERR, "smartd.conf line %d, attributes ids "
			"missing specify attributes ids using -i "
			"option", lineno);
		return -1;
	    }
	    if (FLAG_PARSED >= IDS_PARSED) {
		logging(LOG_ERR,"smartd.conf line %d, Flag args "
			"are more than configured IDS \n", lineno);
		return -1;
	    }

	    num_of_attr = NO_OF_CONFIG_IDS;
	    do {
		arg = strtok(NULL, delim);
		if(!arg) {
		    logging(LOG_ERR, "smartd.conf line %d, parse error "
			    "missing args for -f option", lineno);
		    return -1;
		}

		if (strcasecmp("LOW", arg) == 0) {
		    val = FLAG_LOW;
		} else if (strcasecmp("HIGH", arg) == 0 ) {
		    val = FLAG_HIGH;
		} else {
		    if (*arg == '-') {
			logging(LOG_ERR, "smartd.conf line %d, parse error "
				"missing args for -f option",
				lineno);
		    } else {
			logging(LOG_ERR, "smartd.conf line %d, parse error "
				"flag should be LOW or HIGH for -f option",
				lineno);
		    }
		    return -1;
		}

		for (i = 0; i < NUMBER_ATA_SMART_ATTRIBUTES; i++) {
		    if (dev->ssdthres.thres_entries[i].id ==
			CONFIG_ID(FLAG_PARSED)) {
			dev->ssdthres.thres_entries[i].flag |= FlAG_VALID;
			dev->ssdthres.thres_entries[i].flag |= val;
			SET_ID_STATUS(FLAG_PARSED, FOUND_FLAG);
			FLAG_PARSED++;
			break;
		    }
		}
		num_of_attr--;
	    } while (num_of_attr > 0);

	    SET_PARSED_STATUS(FOUND_FLAG);
	    break;
	default :
	    logging(LOG_ERR, "smartd.conf line %d, parse error "
		    "the option specified -%c is not a valid directive",
		    lineno, sym);
	    return -1;
	    break;
    }

    return 1;
}

/* The function Parses the lines in the config file
 * and return the following values
 * 1: parsed a normal line
 * 0: comment or blankline or config not meant for this device and
    : not a default config
 * -1: Error
 */
static int
parseconfigline(atadevices_t *dev, char *line, int lineno)
{
    const char *delim = " \n\t";
    char *name, *token;
    int i, rc = 0;

    /* get first token: device name. skip parsing line
     * if it a comment or blank line or if config is not
     * for the device and not a default config.
     */
    name = strtok(line, delim);
    if (!name || *name == '#' ||
	/* the config file may contain partial series name */
	((strstr(dev->drive.model, name) == NULL) &&
	 (strcmp(SMARTD_DEFAULT_DEVICE, name)))) {
	return 0;
    }

    /* If the device is in the smartd.conf then update the
     * parse_status and set isdevicessd field.
     */
    dev->ssdthres.isdevicessd = 1;
    SET_PARSED_STATUS(FOUND_DEVICE_NAME);

    /* copy device model name into the ssd_model field */
    strcpy(dev->ssdthres.ssd_model, dev->drive.model);

    /* make a copy of the attribute ids into the smart_ssd_thresholds
     * struct.
     */
    for (i = 0; i < NUMBER_ATA_SMART_ATTRIBUTES; i++) {
	dev->ssdthres.thres_entries[i].id =
	    dev->smartval.vendor_attributes[i].id;
    }

    /* the configuration for the model found update the
     * threshold limits and other data for monitoring
     */
    while ((token = strtok(NULL, delim))) {
	if ((rc = parsetoken(dev, token, lineno)) <= 0) {
	    /*reset the isdevicessd field if the config data is incorrect */
	    dev->ssdthres.isdevicessd = 0;
	    return rc;
	}
    }

    /* Check to see if options for all the IDs are correctly
     * configured
     */
    for (i = 0; (i <= IDS_PARSED); i++) {
	if ((findmissingopt(dev, i, lineno) < 0)) {
	    dev->ssdthres.isdevicessd = 0;
	    rc = -1;
	}
    }
    return rc;
}

/* Read the smartd.conf file and update the vendor specific
 * thershold information for SSD's
 */
void
readconfigfile(void)
{
    /* smartd.conf Path */
    const char *conf_path = SMARTD_CONFIG_FILE;
    static char line[MAXLINELEN+2];
    static char fullline[MAXCONTLINE+1];
    char *lastslash, *comment;
    int lineno, cont, contlineno, slashlines;
    int len, i, rc = 0;
    FILE *conf_fp;
    atadevices_t *dev;
    /* JUNOS begin */ 
    char kern_type[KERN_TYPE_SIZE];
    size_t size = sizeof(kern_type);
    struct stat st;
   
    if (stat(conf_path, &st) != 0) {
        logging(LOG_ERR,"smartd_working.conf not present\n ");
        if (file_copy(SMARTD_CONFIG_FILE, SMARTD_MASTER_CONFIG, 0644)) {
            logging(LOG_ERR,"failed to copy to: %s", SMARTD_CONFIG_FILE);
	}
    }


    if (sysctlbyname(SYSCTL_VM_GUEST, kern_type, &size, NULL, 0) < 0) {
        logging(LOG_ERR, "Could not get sysctl %s\n", SYSCTL_VM_GUEST);
    } else if (strncmp(kern_type, KERN_ON_BARE_METAL, strlen(KERN_ON_BARE_METAL))) {
        logging(LOG_ERR,"calling vmsmartd.sh\n ");
        if (system(PATH_VMSMARTD) != 0) {
           logging(LOG_ERR,"failed to run %s/vmsmartd.sh script", PATH_VMSMARTD);
           return;
        }
    }
    /* JUNOS end */
    /* for all the atadevices scaned check if the thresholds are
     * present in the smartd.conf, if the thresholds are present
     * update the thresholds from the smartd.conf
     */
    for (i = 0; i < numatadevices; i++) {

	dev = &atadevices[i];
	lineno = 1;
	cont = 0;
	contlineno = 0;	
	slashlines = 0;

	/* initalize the parseinfo_t to zero before parsing info
	 * for each device
	 */
	logging(LOG_NOTICE, "%s: %s: ", __FUNCTION__, conf_path);
	memset(&parse_info, 0, sizeof(parseinfo_t));	

	/* initalize the struct ssdthres to zero */
	memset(&dev->ssdthres, 0, sizeof(struct smart_ssd_thresholds));

	/* Open smartd.conf file for reading configuration in text mode */
	if ((conf_fp = fopen(conf_path , "rt")) == NULL) {
	    logging(LOG_ERR, "%s: %s: open: %m", __FUNCTION__, conf_path);
	    return;
	}

	while (fgets(line, MAXLINELEN, conf_fp)) {

	    /* input file line number */
	    contlineno++;

	    if ((comment = strchr(line,'#'))) {
		*comment = '\0';
	    }

	    len = strlen(line);

	    if (len > MAXLINELEN) {
		logging(LOG_WARNING, "%s: %s: line: Exceeds max length ",
			__FUNCTION__, conf_path);
	    }

	    if (cont > MAXCONTLINE) {
		logging(LOG_ERR, "%s: %s: fullline:  Exceeds max length ",
			__FUNCTION__, conf_path);
	    }
	    snprintf(fullline+cont, sizeof(fullline)-cont, "%s" ,line);
	    cont += len;

	    /* is this a continuation line. If so, replace \ by space and look
	     * at next line
	     */
	    if ((lastslash=strrchr(line,'\\')) &&
		!strtok(lastslash+1," \n\t")) {
		*(fullline+(cont-len)+(lastslash-line))=' ';
		slashlines++;
		continue;
	    }

	    /* Get the line no from where the continuation started */
	    lineno = contlineno - slashlines;
	    if ((rc = parseconfigline(dev, fullline, contlineno))) {
		switch (rc) {
		    case -1 :
			logging(LOG_ERR, "%s: %s: Parse: ERROR", __FUNCTION__,
				conf_path);
			logging(LOG_ERR, "Check smartd.conf to "
				"see the usage");
			break;
		    case 1:
			logging(LOG_INFO,"Parsing completed for device "
				"%s", dev->drive.model);
			break;
		    default:
			break;
		}
		break; /* Break the while loop */
	    }
	    slashlines = 0;
	    memset(line, 0, sizeof(line));
	    memset(fullline, 0, sizeof(fullline));
	    cont = 0;
	}

	if (ferror(conf_fp)) {
	    logging(LOG_ERR, "%s: %s: EOF error ", __FUNCTION__, conf_path);
	}

	fclose(conf_fp);
    }
}    	
/* JUNOS end */

/*
 * Scan for SMART capable devices.
 */
static void 
atadevicescan (atadevices_t *devices, char *devicename)
{
    struct ata_params drive;
    int i, fd, numtoscan, support, ret = 0;
    char startdevice[] = "/dev/ad0";
    char *device;
    struct stat stbuf;

    if (devicename == NULL) {
        numtoscan = MAXATADEVICES;
	device = startdevice;
    } else {
        numtoscan = 1;
	device = devicename;
    }

    for(i = 0; i < numtoscan ; i++, device[7]++) {
	fd = 0;
    
	if (stat(device, &stbuf) < 0) {
            if (devicename)
                logging(LOG_ERR, "%s: %s: stat: %s", __FUNCTION__, device, strerror(errno));
	    else
                logging(LOG_DEBUG, "%s: %s: stat: %s", __FUNCTION__, device, strerror(errno));
	    continue;
	}

        if ((fd = open (device , O_RDONLY)) < 0) {
            if (devicename)
                logging(LOG_ERR, "%s: %s: open: %s", __FUNCTION__, device, strerror(errno));
	    else
                logging(LOG_DEBUG, "%s: %s: open: %s", __FUNCTION__, device, strerror(errno));
	    continue;
	}

	logging(LOG_DEBUG, "Opened Device %s", device);

	if (atareadhdidentity(fd, &drive) != 0) {
	    logging(LOG_ERR, "%s: Failed to get information for %s.",
	            __FUNCTION__, device);
	    close(fd);
	    continue;
	}

	support = atasmartsupport(&drive);

	if ((support != 1) && (support != 2) && (support != 255)) {
	    logging(LOG_DEBUG, "%s doesn't support SMART", 
			device);
	    close(fd);
	    continue;
	}

	if ((devicename == NULL) && (driveissupported( &drive )  == FALSE)) {
	    logging(LOG_DEBUG, "%s: not qualified on model %s", SMARTD, 
			drive.model);
	}

	if (((support == 1) || (support == 255)) && (ataenablesmart(fd) != 0)) {
	    logging(LOG_NOTICE, "Unable to enable SMART on %s", 
			device);
	    close(fd);
	    continue;
	}
	    
	devices[numatadevices].fd = fd;
	strlcpy((char *) &devices[numatadevices].devicename, device,
			sizeof(devices[numatadevices].devicename));
	devices[numatadevices].drive = drive;

	ret = atareadsmartvalues (fd, &devices[numatadevices].smartval);
	if (ret != 0) {
	    logging(LOG_ERR, "%s: Failed to get device SMART values", __FUNCTION__);
	    close(fd);
	    continue;
	}

        ret = atareadsmartthresholds(fd, &devices[numatadevices].smartthres);
	if (ret != 0) {
	    logging(LOG_ERR, "%s: Failed to get device SMART thresholds", __FUNCTION__);
	    close(fd);
	    continue;
	}

	devices[numatadevices].selftest = 
			issupportselftest (devices[numatadevices].smartval);
	devices[numatadevices].npassed = 0;
	numatadevices++;

	deviceneedsstandby(&devices[numatadevices]);

	logging(LOG_NOTICE, "Device %s, found and is SMART capable",
			 device);
    }
}

/*
 * Compares SMART values, with the ones already stored.
 */
static void 
atacomparesmartvalues (atadevices_t *device, struct ata_smart_values b)
{
    int i;

    for (i=0; i < NUMBER_ATA_SMART_ATTRIBUTES; i++) {
        if (device->smartval.vendor_attributes[i].id & 
	     b.vendor_attributes[i].id &
             (device->smartval.vendor_attributes[i].normalized != 
	     b.vendor_attributes[i].normalized)) {
	    logging(LOG_DEBUG, "Device: %s, SMART Attribute: %i Changed %x", 
			device->devicename, 
               		device->smartval.vendor_attributes[i].id, 
               		(device->smartval.vendor_attributes[i].normalized - 
			b.vendor_attributes[i].normalized));
        }
    }
}


/*
 * Read SMART values and thresholds. Check and log failure, if any.
 * Also compares values, with the ones stored in the last invocation.
 */
static int 
atacheckdevice(atadevices_t *drive)
{
    struct ata_smart_values tempsmartval;
    struct ata_smart_thresholds tempsmartthres;
    int failed;
    int ret;
    uint disk_smart_status = SMARTD_STATUS_OK;
    uint disk_smart_check = FALSE;
    int len = sizeof(disk_smart_check);

    logging(LOG_DEBUG,"Checking device %s", drive->devicename);

    if ((ret = atareadsmartvalues (drive->fd, &tempsmartval)) != 0) {
	 if (ret == EAGAIN) {
	    return EAGAIN;
	 }
        return -1;
    }

    /* check against the SSD thresholds read from the config file
     * instead of the ones sent by the device.
     * JUNOS begin
     */
    if (drive->ssdthres.isdevicessd) {
	if ((failed = ssdchecksmart(tempsmartval, drive->ssdthres)) != 0) {
	    drive->npassed = 0;
	    if (sysctlbyname(SYSCTL_RE_DISK_SMART_CHECK, &disk_smart_check,
			     &len, NULL, 0) < 0) {
		logging(LOG_ERR, "Could not get sysctl %s\n",
			SYSCTL_RE_DISK_SMART_CHECK);
	    }

	    if(disk_smart_check) {
		disk_smart_status = SMARTD_STATUS_NOK;
		logging (LOG_CRIT, "\nDevice: %s, Failed attribute: %s\n",
			 drive->devicename, atagetsmartattribname(failed));
	    }
	} else {
	    drive->npassed += 1;
	    disk_smart_status = SMARTD_STATUS_OK;
	}

	if (sysctlbyname(SYSCTL_RE_DISK_SMART_STATUS, NULL, 0,
			 &disk_smart_status, sizeof(disk_smart_status)) < 0) {
	    logging(LOG_ERR, "Could not update sysctl %s\n",
		    SYSCTL_RE_DISK_SMART_STATUS);
	}
	/* JUNOS end */
    } else {
	if ((ret = atareadsmartthresholds (drive->fd, &tempsmartthres)) != 0) {
	    if (ret == EAGAIN) {
		return EAGAIN;
	    }
	    return -1;
	}

	if ((failed = atachecksmart(tempsmartval, tempsmartthres)) != 0) {
	    drive->npassed = 0;
	    logging (LOG_CRIT, "\nDevice: %s, Failed attribute: %s\n",
		     drive->devicename, atagetsmartattribname(failed));
	} else {
	    drive->npassed += 1;
	}
    }

    if (drive->npassed == checksinamonth) {
	drive->npassed = 0;
	logging(LOG_DEBUG,"Device %s passed SMART checks in the past 30 days", 
		    drive->devicename);
    } else {
	logging(LOG_DEBUG,"Device %s passed SMART check", drive->devicename);
    }

    atacomparesmartvalues (drive , tempsmartval);
        
    /* Copy the values back to the structure */
    drive->smartval = tempsmartval;

    /* JUNOS begin
     * If the device is SSD don't update the thrsholds
     */
    if (!drive->ssdthres.isdevicessd) {
	drive->smartthres = tempsmartthres;
    }
    /* JUNOS end */

    return 0;
}

static __inline__ void
sby_trace_spin(ad_arm_t *armi)
{
    const char *reason;
    pid_t pid;
    const char *procname;
    char tbuf[TIME_BUF_SZ];
    
    if (armi->spinup_time_valid) {
        procname = armi->unarm_proc;
        pid = armi->unarm_pid;
        reason = reason2str(armi->unarm_reason);
        if (!*procname) {
            procname = "(null)";
            pid = -2;
        }
        if (spindown_time_valid) { 
            ctime_r((time_t *)&spindown_time_tv.tv_sec, tbuf);
            sby_trace("down %ld.%03lds, spun down %s",
                spindown_time_sec, spindown_time_msec, (tbuf+4));
            spindown_time_valid = 0;
        }
        sby_trace("up %ld.%03lds, %s, %s, %d \n",
            (long int)armi->spinup_time.tv_sec, (armi->spinup_time.tv_usec/1000 +
          ((armi->spinup_time.tv_usec%1000 >= 500) ? 1 : 0)),
            reason, procname, pid);
        armi->spinup_time_valid = 0;
    }
    if (armi->spindown_time_valid) {
        /* 
         * Logging the "down" message for reporting the downtime,
         * to the smartd trace file brings the disk out of standby 
         * too early. So, don't disturb the disk now, but do make 
         * a note of it here to do the actual logging later when 
         * the disk is in fact out of standby already. 
         */ 
        spindown_time_valid = 1;
        spindown_time_sec  = armi->spindown_time.tv_sec; 
        spindown_time_msec = armi->spindown_time.tv_usec/1000 +
                           ((armi->spindown_time.tv_usec%1000 >= 500) ? 1 : 0);
        gettimeofday(&spindown_time_tv, NULL);
        armi->spindown_time_valid = 0;
    }
}

static void
sby_trace(const char *fmt, ...)
{
    va_list ap;
    int size = sizeof(tracebuf);
    int len = 0;
    int ret = 0;
    struct timeval tv;

    if (tracefd >= 0) {
        gettimeofday(&tv, NULL);
        len = syslogtime(&tv, tracebuf, size, LOG_YEAR);
        len += snprintf(tracebuf+len, size-len, "smartd[%d]: ", smartdpid);
        va_start(ap, fmt);
        len += vsnprintf(tracebuf+len, size-len, fmt, ap);
        va_end(ap);
        if (len >= size) {
            tracebuf[size-1] = '\0';
            len = size;
        }
        if ((ret = write(tracefd, tracebuf, len)) < 0) {
            logging(LOG_ERR, "Error writing to file %s, fd %d: %s\n",
                    SMARTD_TRACE_FILE, tracefd, strerror(errno));
            va_start(ap, fmt);
            vsyslog(LOG_INFO, fmt, ap);
            va_end(ap);
        }
    }

    if (standbyloglevel) {
        va_start(ap, fmt);
        vsyslog(standbyloglevel, fmt, ap);
        va_end(ap);
    }
}

const char *
reason2str(char reason)
{
    static char buf[8];

    switch (reason) {
        case AD_REASON_NOTSET:
            return "notset";
            break;
        case AD_REASON_READ:
            return "read";
            break;
        case AD_REASON_WRITE:
            return "write";
            break;
        case AD_REASON_IOCTL:
            return "ioctl";
            break;
        case AD_REASON_TIME:
            return "time";
            break;
        case AD_REASON_TIMEOUT:
            return "timeout";
            break;
        case AD_REASON_RUSH:
            return "rush";
            break;
        case AD_REASON_OTHER:
            return "other";
            break;
        default:
            snprintf(buf, sizeof(buf), "0x%2x", reason);
            return buf;
            break;
    }
}

static int
deviceneedsstandby (atadevices_t *drive)
{
    int ret = FALSE;
    int i;

    if (!drive) {
        return ret;
    }

    for (i = 0; atasbytab[i].disk_model; i++) {
        if (strcmp(drive->drive.model, atasbytab[i].disk_model)) {
            continue;
        }
        sby_dlog("Found disk %s \n", drive->drive.model);
        ret = TRUE;
        break;
    }
    drive->sby_table_index = i;
    return ret;
}

void
readstandbyconfig (void)
{
    int no_sby_periodic = FALSE;
    int dumb_standby;
    int sby_type   = SBY_ADAPTIVE_ALWAYS;
    int sby_enable = FALSE;
    int sby_interval = 0;
    int sby_loglevel = 0;
    time_t sby_summary = 0;
    size_t len = sizeof(no_sby_periodic);
    size_t dlen = sizeof(disk_inactivity_timer);
    int i = 0;
    uint j = 0;
    uint old_tmr      = disk_inactivity_timer;
    atadevices_t *dev = NULL;

    /* close log file and reopen it */	
    if (tracefd >= 0) {
	close(tracefd);
    }
    if ((tracefd = open(SMARTD_TRACE_FILE, O_CREAT|O_APPEND|O_RDWR,
                        S_IRWXU|S_IRGRP|S_IROTH)) < 0) {
        logging(LOG_ERR, "Error opening %s : %s\n", SMARTD_TRACE_FILE, strerror(errno));
    }

    if (sysctlbyname(SYSCTL_RE_DISK_INACTIVITY_TIMER, &disk_inactivity_timer,
                     &dlen, NULL, 0) < 0) {
	logging(LOG_ERR, "Could not get sysctl %s\n",
		SYSCTL_RE_DISK_INACTIVITY_TIMER);
	disk_inactivity_timer = SMARTD_DISK_INACT_TMR;
    } else {
	/* In seconds */
	disk_inactivity_timer = disk_inactivity_timer * 60;
        sby_trace("Disk inactivity timer configured to %d sec.\n",
                 disk_inactivity_timer);
    }


    /* Check if standby debug loglevel was set through a sysctl. We
     * do this check if it was not set already through smartd's
     * command line -Jl option */
    len = sizeof(sby_loglevel);
    if (!cmd_loglevel) {
        if (sysctlbyname(SYSCTL_RE_STANDBY_LOGLEVEL, &sby_loglevel,
                         &len, NULL, 0) < 0) {
            if (errno != ENOENT) {
                logging(LOG_ERR, "Could not get sysctl %s: %s\n", 
                        SYSCTL_RE_STANDBY_LOGLEVEL, strerror(errno));
            }
            sby_loglevel = 0;
        } else {
            /* Sanity check. If level is incorrect then we ignore it */
            if (sby_loglevel > LOG_DEBUG || sby_loglevel < 0) {
                logging(LOG_ERR, "Invalid value %d for sysctl %s\n", 
                        sby_loglevel, SYSCTL_RE_STANDBY_LOGLEVEL);
                sby_loglevel = 0;
            }
        }
    } else {
        sby_loglevel = cmd_loglevel;
    }
    /* We turn it on so that trace messages in this function are
     * also printed in syslog */
    standbyloglevel = sby_loglevel;

    /* Set the summary period */
    if (!cmd_summaryperiod) {
        sby_summary = SMARTD_DAILY_CHECK;
    } else {
        sby_summary = cmd_summaryperiod;
    }

    /* If a standby interval was forced on command line then that
     * overrides any sysctl settings. Standby is always enabled
     * in this case.
     */
    if (cmd_standbyinterval) {
        sby_interval = cmd_standbyinterval;
        sby_enable = TRUE;
        goto done;
    }
    /* 
     * Check to see if we need to switch to dumb(perodic standby)
     * from adaptive
     */
    len = sizeof(dumb_standby);
    if (sysctlbyname(SYSCTL_RE_DISK_STANDBY_PERIODIC, &dumb_standby,
                     &len, NULL, 0) < 0) {
        if (errno != ENOENT) {
            logging(LOG_ERR, "Could not get sysctl %s: %s\n", 
                    SYSCTL_RE_DISK_NO_STANDBY, strerror(errno));
        }
        /* On error we assume that periodic standby has not been disabled */
        sby_type = SBY_ADAPTIVE_ALWAYS;
    }

    if (TRUE == dumb_standby) {
	sby_type = SBY_PERIODIC_ALWAYS;
    }

    sby_trace("Standby operation set to %s\n",
	      (sby_type == SBY_PERIODIC_ALWAYS) ? "forced periodic": "adaptive periodic");

    /* Check to see if the standby has been disabled through CLI */
    len = sizeof(no_sby_periodic);
    if (sysctlbyname(SYSCTL_RE_DISK_NO_STANDBY, &no_sby_periodic,
                     &len, NULL, 0) < 0) {
        if (errno != ENOENT) {
            logging(LOG_ERR, "Could not get sysctl %s: %s\n", 
                    SYSCTL_RE_DISK_NO_STANDBY, strerror(errno));
        }
        /* On error we assume that periodic standby has not been disabled */
        no_sby_periodic = FALSE;
        sby_enable = TRUE;
    }

    if (TRUE == no_sby_periodic) {
        if (standbyinterval) {
            sby_trace("Request disabled\n");
        }
        sby_enable = FALSE;
        goto done;
    }

    /* Check if a specific interval has been specified through CLI */
    len = sizeof(sby_interval);
    if (sysctlbyname(SYSCTL_RE_DISK_STANDBY_INTERVAL, &sby_interval,
                     &len, NULL, 0) < 0) {
        if (errno != ENOENT) {
            logging(LOG_ERR, "Could not get sysctl %s: %s, default "
            "interval %d will be used\n", SYSCTL_RE_DISK_STANDBY_INTERVAL, strerror(errno),
            SMARTD_STANDBY_INTERVAL_DEFAULT);
        }
        sby_interval = 0;
        sby_enable = FALSE;
    }

    /* Standby interval has been specified in CLI */
    if (sby_interval) {
        sby_enable = TRUE;
    }

    if (sby_interval > SMARTD_STANDBY_INTERVAL_MAX) {
        logging(LOG_ERR, "Interval has invalid value %d. Should be "
                "less than %d\n", sby_interval, SMARTD_STANDBY_INTERVAL_MAX);
        logging(LOG_ERR, "Setting interval to default value of %d "
                "minutes\n", SMARTD_STANDBY_INTERVAL_DEFAULT);
        sby_interval = SMARTD_STANDBY_INTERVAL_DEFAULT;
    }

    /* If no interval has been set but periodic standbys are enabled
     * then default to issuing a standby once every 4 hours */
    if (!sby_interval) {
        sby_interval = SMARTD_STANDBY_INTERVAL_DEFAULT;
    }

    /* convert to seconds */
    sby_interval = 60 * sby_interval;

done:
    /* If the config change is relevant for individual devices then wake
     * up their standby_request() functions immediately to apply the change
     */
    if (old_tmr != disk_inactivity_timer) {
	for (i = 0; i < numatadevices; i++) {
            dev = &atadevices[i];
	    if (ataissueidle(dev->fd, disk_inactivity_timer) < 0) {
		logging(LOG_ERR, "Device %s: unable to issue idle %d\n",
                    dev->devicename,  disk_inactivity_timer);
	    } else {
		sby_trace("Device %s: inactivity timer enabled to %d \n",
		    dev->devicename, disk_inactivity_timer);
	    }
	}
    }

    for (j = 0; j < sizeof(atasbytab)/sizeof(sby_table_t); j++) {
	atasbytab[i].sby_type = sby_type;
    }

    if (standbyinterval != sby_interval || cli_standbyenable != sby_enable) {
        for (i = 0; i < numatadevices; i++) {
            dev = &atadevices[i];
            if (!evTestID(dev->standby_evtid)) {
                continue;
            }
            if (evResetTimer(smartd_ctx, dev->standby_evtid, standby_request,
                        dev, evConsTime(0,0), evConsTime(0, 0)) < 0) {
                logging(LOG_EMERG, "evReSetTimer(config) failed : %s\n", strerror(errno));
                /* not reached */
            }
        }
    }
    standbyinterval = sby_interval;
    standbysummary = sby_summary;
    standbyloglevel = sby_loglevel;
    cli_standbyenable = sby_enable;
    /* Check to see if power check has been enabled through CLI */
    len = sizeof(periodic_power_check);
    if (sysctlbyname(SYSCTL_RE_DISK_PERIODIC_POWER_CHECK,
                     &periodic_power_check, &len, NULL, 0) < 0) {
        if (errno != ENOENT) {
            logging(LOG_ERR, "Could not get sysctl %s: %s\n",
                    SYSCTL_RE_DISK_PERIODIC_POWER_CHECK, strerror(errno));
        }
        /* On error we assume that power check has been disabled */
        periodic_power_check = FALSE;
    }

    /*
     * XXX: Remove the long typecast for size_t variables once %z format is
     *      recongized by our compilers
     */
    sby_trace("New config: Periodic power check %s, "
            "Periodic Standby %s: Interval %ld secs, Summary "
            "interval %ld secs, loglevel %d\n",
	    periodic_power_check ? "enabled": "disabled",
            no_sby_periodic ? "disabled" : "enabled", (long)standbyinterval,
            (long)standbysummary, standbyloglevel);
}

static int
sby_is_enabled(sby_table_t *sby)
{
    /* Standby type says standby is not needed */
    if (SBY_NOTNEEDED == sby->sby_type) {
        return 0;
    }

    /* Standby is explicitly disabled in CLI */
    if (!standbyinterval) {
        return 0;
    }

    /* Standby is explicitly enabled in CLI */
    if (cli_standbyenable) {
        return 1;
    }

    /* Standby is not explicitly enabled in CLI. And the standby type
     * needs such explicit enabling
     */
    if ((SBY_PERIODIC_IF_ENABLED == sby->sby_type) ||
        (SBY_ADAPTIVE_IF_ENABLED == sby->sby_type)) {
        return 0;
    }
    
    /* Standby is implicitly enabled and the standby type only requires
     * such implicit enabling
     */
    if ((SBY_PERIODIC_ALWAYS == sby->sby_type) ||
        (SBY_ADAPTIVE_ALWAYS == sby->sby_type)) {
        return 1;
    }

    /* Catch all. If new standby types are introduced then we should
     * introduce them in this function.
     */
    assert(0);

    return 0;
}

static void
smart_check (evContext ctx, void *uap,
             struct timespec due __unused,
             struct timespec inter __unused)
{
    int ret = 0;
    atadevices_t *dev = (atadevices_t *) uap;

    /* If disk is in standby and less than 5 minutes have
     * elapsed while it was in standby then skip one round
     * of smart checks.
     *
     * If disk has been in standby greater than 5 minutes or
     * we already skipped checks once then we execute the smart checks.
     * This could potentially bring the disk out of standby.
     */
    if (!dev->skipped_smart_once && dev->in_standby &&
        dev->sby_curr_accum_time <= SMARTD_STANDBYTIME_MIN) {
        dev->skipped_smart_once = TRUE;
        goto done;
    }

    dev->skipped_smart_once = FALSE;

    if ((ret = atacheckdevice (dev)) != 0) {
        if (ret == EAGAIN) {
            logging(LOG_ERR, "%s: Device %s is busy, could not run "
                "atacheckdevice", dev->devicename,
                __FUNCTION__);
        } else {
            logging(LOG_ERR, "%s: Device %s, non zero return from "
                "atacheckdevice", dev->devicename,
                __FUNCTION__);
        }
    }

done:
    /* Rearm the timer. We will run every checktime seconds */
    if (evSetTimer(ctx, smart_check, dev,
            evAddTime(evNowTime(), evConsTime(checktime, 0)),
            evConsTime(0, 0), NULL) < 0) {
        logging(LOG_EMERG, "evSetTimer(smart_check) failed : %s\n", strerror(errno));
        /* not reached */
    }
    return;
}

static void
power_check (evContext ctx, void *uap,
             struct timespec due __unused,
             struct timespec inter __unused)
{
    atadevices_t *dev = (atadevices_t *) uap;
    sby_table_t  *sby = &atasbytab[dev->sby_table_index];
    time_t power_time = SMARTD_POWER_CHECK_1;
    time_t sby_power_time = 0;
    char power_mode = ATA_POWERMODE_ACTIVE_OR_IDLE;
    char standby_armed = FALSE;
    time_t sby_time = 0;


    /* Check the power mode first and then the standby arm request */
    power_mode = atacheckpowermode(dev->fd, 
	(unsigned char)periodic_power_check);

    /* First lets deal with the arm standby request */
    if (dev->sby_armed) {
        ad_arm_t    arm_info;

        standby_armed = atastandbyarmget(dev->fd, &arm_info);
        if (standby_armed < 0) {
            standby_armed = 0;
        } else {
            sby_trace_spin(&arm_info);
        }

        /* It is not possible for standby_armed to be TRUE along with drive
         * being in standby. Flag an error message if this happens and
         * unarm the standby request
         */
        if (standby_armed && (power_mode == ATA_POWERMODE_STANDBY)) {
            logging(LOG_ERR, "%s: %s: Found device inactive and "
                    "arm request is also TRUE, disarming arm request\n",
                    dev->devicename, __FUNCTION__);
            atastandbyarmset(dev->fd, FALSE);
            dev->sby_armed = FALSE;
        }
        else if (!standby_armed) {
            dev->sby_armed = FALSE;
            /* Standby request has been unarmed. Means that the driver
             * issued the standby request. There is a slim chance that
             * the drive actually went to standby on its own before
             * the driver issued the standby. But since we cannot
             * distinguish this case we assume that the source is
             * always smartd whenever standby request is unarmed.
             */
            assert(!dev->in_standby && (dev->sby_source == SBY_SRC_NOTSET));
            dev->sby_source = SBY_SRC_SMARTD;
            dev->in_standby = TRUE;
            dev->sby_curr_accum_time = 0;
            dev->sby_curr_start_time = dev->last_power_check_time = time(NULL);
            dev->skipped_smart_once = FALSE;
            sby_dlog("Device %s:%s: adaptive standby was unarmed by kernel, "
                 "assuming that disk in standby due to smartd\n",
                 dev->devicename, __FUNCTION__);
        }
        /* Adaptive Standby has been armed. Check whether disk has actually
         * gone to standby every 1s
         */
        sby_power_time = SMARTD_POWER_CHECK_3;
    }

    if (power_mode == ATA_POWERMODE_STANDBY) {
        /* Disk went to standby on its own due to its standby timer */
        if (!dev->in_standby) {
            assert(dev->sby_source == SBY_SRC_NOTSET);
            assert(!dev->sby_armed);
            dev->sby_source = SBY_SRC_DISK;
            dev->in_standby = TRUE;
            dev->sby_curr_accum_time = 0;
            dev->sby_curr_start_time = dev->last_power_check_time = time(NULL);
            dev->skipped_smart_once = FALSE;
        } else {
            /* Disk was in standby and continues to remain there */
            assert(dev->sby_source != SBY_SRC_NOTSET);
            dev->last_power_check_time = time(NULL);
        }
        /* Disk is in standby, check power mode once every 2s */
        power_time = SMARTD_POWER_CHECK_2;
    } else {
        /* Did disk come out of standby ? Log it */
        if (dev->in_standby) {
            ad_arm_t    arm_info;
            const char *reason;
            pid_t pid;
            const char *procname;

            assert(dev->sby_source != SBY_SRC_NOTSET);

            sby_time = (dev->last_power_check_time - dev->sby_curr_start_time);


            if (atastandbyarmget(dev->fd, &arm_info) >= 0) {
                sby_trace_spin(&arm_info);
                procname = arm_info.unarm_proc;
                pid = arm_info.unarm_pid;
                reason = reason2str(arm_info.unarm_reason);
                if (!*procname) {
                    procname = "(null)";
                    pid = -2;
                }
            } else {
                procname = "(null)";
                pid = -1;
                reason = reason2str(AD_REASON_NOTSET);
            }

            /*
             * XXX: Remove the u_long typecast for size_t variables once 
	     *      %z format is recongized by our compilers
             */
            sby_trace("Device %s, %s, %lds, %s, %s, %d\n", dev->devicename,
                (dev->sby_source == SBY_SRC_DISK) ? "disk":"smartd",
                (long)sby_time, reason, procname, pid);

            dev->sby_daily_accum_time += sby_time;

            /* For adaptive standby, standby-time below a certain threshold
             * does not count for one of the accumulation counts
             */
            if (sby_is_adaptive(sby)) {
                if (sby_time >= sby->accum_time_per_sby_min) {
                    dev->sby_valid_daily_accum_time += sby_time;
                }
            }
            else if (sby_is_periodic(sby)) {
                dev->sby_valid_daily_accum_time += sby_time;
            } 
            /* Record the largest continuous standby time */
            if (sby_time > dev->sby_continuous_time_max) {
                dev->sby_continuous_time_max = sby_time;
            }
            dev->sby_daily_num++;
            dev->in_standby = FALSE;
            dev->sby_curr_accum_time = dev->sby_curr_start_time = 0;
            dev->last_power_check_time = 0;
            dev->sby_source = SBY_SRC_NOTSET;
        } else {
            assert(dev->sby_source == SBY_SRC_NOTSET);
        }
        /* Disk is not in standby, check power mode every 15s */
        power_time = SMARTD_POWER_CHECK_1;
    }
    /* 
     * If adaptive standby has been armed then we check status more
     * aggressively
     */
    if (sby_power_time) {
        power_time = sby_power_time;
    }
    /* Rearm power check */
    if (evSetTimer(ctx, power_check, dev,
            evAddTime(evNowTime(), evConsTime(power_time, 0)),
            evConsTime(0,0),  &dev->power_evtid) < 0) {
        logging(LOG_EMERG, "evSetTimer(power_check) failed : %s\n", strerror(errno));
        /* not reached */
    }
}

static void
standby_request (evContext ctx, void *uap,
                 struct timespec due __unused,
                 struct timespec inter __unused)
{
    atadevices_t *dev = (atadevices_t *) uap;
    sby_table_t  *sby = &atasbytab[dev->sby_table_index];

    if (!sby_is_enabled(sby)) {
        if (dev->sby_interval) {
            sby_dlog("%s: Disabling %s\n", dev->devicename, __FUNCTION__);
            dev->sby_interval = 0;

#if 0
            /* Enable write caching unless forced to be off */
            if (!sby->disable_wcache) {
                if (atawritecacheset(dev->fd, TRUE) < 0) {
                    logging(LOG_ERR,"Device %s: unable to enable write cache\n",
                        dev->devicename);
                } else {
                    sby_dlog("Device %s:%s: write cached enabled \n",
                            dev->devicename, __FUNCTION__);
                }
            }
#endif 
#if 0
	    /* Timer should be set only once */		
            /* Enable disk inactivity timer unless permanently disabled */
            if (sby->inactivity_timer) {
                if (ataissueidle(dev->fd, sby->inactivity_timer) < 0) {
                    logging(LOG_ERR, "Device %s: unable to issue idle %d\n",
                        dev->devicename, sby->inactivity_timer);
                } else {
                    sby_dlog("Device %s:%s: inactivity timer enabled \n",
                            dev->devicename, __FUNCTION__);
                }
            }
#endif

        }

        if (evSetTimer(ctx, standby_request, dev,
            evAddTime(evNowTime(), evConsTime(SMARTD_CONFIG_APPLY_TIME, 0)),
            evConsTime(0,0), &dev->standby_evtid) < 0) {

            logging(LOG_EMERG, "evSetTimer(inactive_request) failed : %s\n", strerror(errno));
            /* not reached */
        }
        return;
    }

    assert(sby_is_adaptive(sby) || sby_is_periodic(sby));

    /*
     * XXX: Remove the u_long typecast for size_t variables once %z format is
     *      recongized by our compilers
     */
    if (!dev->sby_interval) {
        sby_dlog("%s: Enabling %s standby interval %ld seconds\n",
                dev->devicename, __FUNCTION__, (long)standbyinterval);
        dev->sby_interval = standbyinterval;
#if 0
        /* Disable write caching unless forced to be off. If it is forced
         * to be off always then it would have been done in armtimers()
         */
        if (!sby->disable_wcache) {
            if (atawritecacheset(dev->fd, FALSE) < 0) {
                logging(LOG_ERR, "Device %s: unable to disable write cache\n",
                    dev->devicename);
            } else {
                sby_dlog("Device %s:%s: write cache disabled\n",
                        dev->devicename, __FUNCTION__);
            }
        }
#endif 
#if 0
	/* Need to set only once */
        /* Disable disk inactivity timer unless already permanently disabled */
        if (sby->inactivity_timer) {
            if (ataissueidle(dev->fd, 0) < 0) {
                logging(LOG_ERR, "Device %s: unable to issue idle %d\n",
                    dev->devicename, sby->inactivity_timer);
            } else {
                sby_dlog("Device %s:%s: inactivity timer disabled\n",
                        dev->devicename, __FUNCTION__);
            }
        }
#endif 
    }

    /*
     * XXX: Remove the u_long typecast for size_t variables once %z format is
     *      recongized by our compilers
     */
    if (dev->sby_interval != standbyinterval) {
        sby_dlog("%s: %s: interval changed from %ld to %ld "
            "seconds\n", dev->devicename, __FUNCTION__, (long)dev->sby_interval,
            (long)standbyinterval);
        dev->sby_interval = standbyinterval;
    }

    /* Adaptive Algorithm for standby. Kernel will try to buffer asynchronous
     * writes for a longer period after standby is issued.
     */
    if (sby_is_adaptive(sby)) {
        /* Drive is still armed to issue standby from a previous run of
         * this function. We do nothing.
         */
        if (dev->sby_armed) {
            goto done;
        }

        /* we were asked to accumulate only standby times > 60s and were asked
         * to stop when such accumulated times for the day reach the daily
         * required quota.
         */
        if ((STOP_AT_ACCUM_LIMIT == sby->stop_at_limit) &&
            (dev->sby_valid_daily_accum_time >= sby->accum_time_max)) {
            goto done;
        }

        /*
         * We were asked to continue till we hit at least one continuous
         * interval of the daily required quota.
         * 
         * If we were given standby interval through command line then go ahead
         * with the standby even if we have completed the maximum 
         * continuous standby for the day. This is only for testing purpose.
         */
        if ((STOP_AT_CONT_MAX_LIMIT == sby->stop_at_limit) && 
            (dev->sby_continuous_time_max >= sby->accum_time_max)
            && !cmd_standbyinterval) {
            goto done;
        }

        if (!dev->in_standby) {
            assert(dev->sby_source == SBY_SRC_NOTSET);
            /* Arm the standby */
            if (atastandbyarmset (dev->fd, TRUE) < 0) {
                logging(LOG_ERR, "%s: Error: atastandbyarmset(TRUE):"
                    " %s", __FUNCTION__, strerror(errno));
            } else {
                dev->sby_armed = TRUE;
                /* Speed up the power check */
                assert(evTestID(dev->power_evtid));
                if (evResetTimer(ctx, dev->power_evtid, power_check, dev,
                                 evConsTime(0, 0), evConsTime(0, 0)) < 0) {
                    logging(LOG_EMERG,
                        "evResetTimer(inactive_request) failed : %s\n", strerror(errno));
                    /* not reached */
                }
            }
        } else {
            /* Do nothing if disk is already in standby */
            assert(dev->sby_source != SBY_SRC_NOTSET);
        }

    } else if (sby_is_periodic(sby)) {
        /* Forced periodic standby. Kernel just issues the standby. Does
         * not buffer async writes */
        if (!dev->in_standby) {
            assert(dev->sby_source == SBY_SRC_NOTSET);
            if (ataissuestandbyimmediate (dev->fd) < 0) {
                logging(LOG_ERR, "%s: Error: ataissuestandbyimmediate:"
                    " %s", __FUNCTION__, strerror(errno));
            } else {
                dev->sby_source = SBY_SRC_SMARTD;
                dev->in_standby = TRUE;
                dev->sby_armed  = 0;
                dev->sby_curr_start_time =
                                        dev->last_power_check_time = time(NULL);
                dev->sby_curr_accum_time = 0;
                dev->skipped_smart_once = FALSE;
                sby_dlog("Device %s:%s: disk went into standby due to smartd\n",
                      dev->devicename, __FUNCTION__);

                /* Speed up the power check */
                assert(evTestID(dev->power_evtid));
                if (evResetTimer(ctx, dev->power_evtid, power_check, dev,
                                 evConsTime(0, 0), evConsTime(0, 0)) < 0) {
                    logging(LOG_EMERG,
                        "evResetTimer(inactive_request) failed : %s\n", strerror(errno));
                    /* not reached */
                }
            }
        } else {
            /* Do nothing if disk is already in standby */
            assert(dev->sby_source != SBY_SRC_NOTSET);
        }
    }

done:
    if (evSetTimer(ctx, standby_request, dev,
        evAddTime(evNowTime(), evConsTime(dev->sby_interval, 0)),
        evConsTime(0,0), &dev->standby_evtid) < 0) {

        logging(LOG_EMERG, "evSetTimer(inactive_request) failed : %s\n", strerror(errno));
        /* not reached */
    }
}

static void
daily_check (evContext ctx, void *uap,
             struct timespec due __unused,
             struct timespec inter __unused)
{
    int i;
    atadevices_t *atadevs = (atadevices_t *) uap;
    atadevices_t *dev = NULL;
    sby_table_t  *sby = NULL;

    for (i = 0; i < numatadevices; i++) {
        dev = &atadevs[i];
        sby = &atasbytab[dev->sby_table_index];
        if (sby->sby_type == SBY_NOTNEEDED) {
            continue;
        }

        /*
         * XXX: Remove the u_long typecast for size_t variables once
	 *      %z format is recongized by our compilers
         */
        if (sby_is_periodic(sby)) {
            sby_trace("Device %s: inactive (%ld,%ld,%ld,%ld,%u)s\n",
            dev->devicename, (long)dev->sby_daily_accum_time, 0L,
            (long)dev->sby_continuous_time_max, 0L, dev->sby_daily_num);
        }
        if (sby_is_adaptive(sby)) {
            sby_trace("Device %s: inactive (%" PRIdTM ",%" PRIdTM ",%" PRIdTM
	    ",%" PRIdTM ",%u)s\n",
            dev->devicename, dev->sby_daily_accum_time,
            dev->sby_valid_daily_accum_time, dev->sby_continuous_time_max,
            sby->accum_time_per_sby_min, dev->sby_daily_num);
        }
        dev->sby_continuous_time_max = 0;
        dev->sby_daily_accum_time = 0;
        dev->sby_valid_daily_accum_time = 0;
        dev->sby_daily_num = 0;
    }

    /* Rearm the timer */
    if (evSetTimer(ctx, daily_check, atadevs,
            evAddTime(evNowTime(), evConsTime(standbysummary, 0)),
            evConsTime(0,0), NULL) < 0) {
        logging(LOG_EMERG, "evSetTimer(daily_check) failed : %s\n", strerror(errno));
        /* not reached */
    }
    return;
}

static void
armtimers (atadevices_t *atadevs)
{
    int i = 0;
    atadevices_t *dev = NULL;
    sby_table_t  *sby = NULL;
    time_t      sby_period = 0;
    time_t      power_period = 0;


    for (i = 0; i < numatadevices; i++) {
        dev = &atadevs[i];

        assert(dev->sby_table_index < sizeof(atasbytab)/sizeof(atasbytab[0]));

        dev->in_standby = FALSE;
        dev->sby_source = SBY_SRC_NOTSET;
        dev->sby_interval = 0;
        dev->sby_curr_start_time = 0;
        dev->last_power_check_time = 0;
        dev->sby_curr_accum_time = 0;
        dev->sby_daily_accum_time = 0;
        dev->sby_valid_daily_accum_time = 0;
        dev->sby_continuous_time_max = 0;
        dev->skipped_smart_once = FALSE;
        dev->sby_daily_num = 0;
        dev->sby_armed = FALSE;
        evInitID(&dev->power_evtid);
        evInitID(&dev->standby_evtid);

        sby = &atasbytab[dev->sby_table_index];

#if 0
	/* ata driver aleady sets this */
        /* Disable write caching for good if needed for the device */
        if (sby->disable_wcache) {
            if (atawritecacheset(dev->fd, FALSE) < 0) {
                logging(LOG_ERR, "Device %s: unable to disable write cache\n",
                    dev->devicename);
            } else {
                sby_dlog("Device %s: write cached disabled \n",
                        dev->devicename);
            }
        }
#endif
        /* Set timer to cofigure value  */
	if (ataissueidle(dev->fd, disk_inactivity_timer) < 0) {
	    logging(LOG_ERR, "Device %s: unable to issue idle %d\n",
                    dev->devicename,  disk_inactivity_timer);
	} else {
	    sby_trace("Device %s: inactivity timer set to %d \n",
		dev->devicename, disk_inactivity_timer);
	}

        /* Enable periodic smart check. We actually start off the first
         * one immediately using evConsTime(0,0). Subsequent onces
         * are fired once every 30 minutes
         */
        if (evSetTimer(smartd_ctx, smart_check, dev, evConsTime(0,0),
                       evConsTime(0,0), NULL) < 0) {
            logging(LOG_EMERG, "evSetTimer(smart_check) failed : %s\n", strerror(errno));
            /* not reached */
        }

        /* Enable periodic power check to see if drive has gone to standby
         * Offset it by 5 seconds so that it doesn't start exactly at the
         * same time as the smart checks */
        power_period = SMARTD_POWER_CHECK_1 + SMARTD_POWER_CHECK_OFFSET;
        if (evSetTimer(smartd_ctx, power_check, dev,
                evAddTime(evNowTime(), evConsTime(power_period, 0)),
                evConsTime(0,0), &dev->power_evtid) < 0) {
            logging(LOG_EMERG, "evSetTimer(power_check) failed : %s\n", strerror(errno));
            /* not reached */
        }

        switch (sby->sby_type) {
            case SBY_NOTNEEDED:
                continue;
                break;

            default:
                break;
        }

        /* Has periodic standby been enabled ? */
        if (sby_is_enabled(sby)) {
            sby_period = standbyinterval;
            /* Offset it by 15 minutes so that the regular 4 hour periodic 
             * standfy falls midway between the regular half hour smart checks
             *
             * Do this only if we are not in the test mode.
             */
            if (!cmd_standbyinterval) {
                sby_period += SMARTD_STANDBY_INTERVAL_OFFSET;
            }
        } else {
            sby_period = SMARTD_CONFIG_APPLY_TIME;
        }

        /* Enable periodic standby */
        if (evSetTimer(smartd_ctx, standby_request, dev,
                evAddTime(evNowTime(), evConsTime(sby_period, 0)),
                evConsTime(0,0), &dev->standby_evtid) < 0) {
            logging(LOG_EMERG, "evSetTimer(inactive) failed : %s\n", strerror(errno));
            /* not reached */
        }

        /* Enable Daily Resetting of Accumulated standby times */
        if (evSetTimer(smartd_ctx, daily_check, atadevs,
                evAddTime(evNowTime(), evConsTime(standbysummary, 0)),
                evConsTime(0,0), NULL) < 0) {
            logging(LOG_EMERG, "evSetTimer(daily_check) failed : %s\n", strerror(errno));
            /* not reached */
        }

    }
    return;
}

/*
 * Check for SMART failure on all the scanned devices.
 */
static void 
checkdevices ( atadevices_t *atadevs)
{
    int ret = 0;
    
    checksinamonth = 3600 * 24 * 30 / checktime;	/* checks in 30 days */
    if (checksinamonth == 0)
	checksinamonth = 1;

    /* JUNOS begin */
    if (file_copy(SMARTD_CONFIG_FILE, SMARTD_MASTER_CONFIG, 0644)) {
	logging(LOG_ERR,"failed to copy to: %s", SMARTD_CONFIG_FILE);
    }

    readconfigfile();
    /* JUNOS end */

    armtimers(atadevs);
    ret = evMainLoop(smartd_ctx);
    logging(LOG_EMERG, "evMainLoop exit (%d): %s\n", ret, strerror(errno));
    /* not reached */
    exit(1);
}

/*
 * Scan for SMART capable devices and check for SMART failure.
 */
int 
daemonloop (char *device)
{
    atadevices_t *atadevicesptr;
    
    numatadevices = 0;
    atadevicesptr = atadevices;

#ifdef HAVE_SIL3132_CHIPSET
    /*
     * According to Silicon Image (vendor of SATA chip sil3132), add 60 seconds
     * delay on getting first SMART data to avoid chksum error.
     * This delay will NOT bring any side affect on disk monitorring.
     *
     */
    sleep(60);
#endif

    /* Read sysctls for issuing periodic disk standby */
    readstandbyconfig();
    sby_trace("Starting S.M.A.R.T \n");

    /* 
     * Scan for SMART capable devices, till one is found. If user specified
     * a device on the command line and its not SMART capable, flag it as an 
     * error.
     */
    while (numatadevices == 0) {
	atadevicescan (atadevicesptr, device); 
	if (numatadevices != 0) {
	    break;
	}
	if (device != NULL) {
	    return ENOENT;
	} else {
	    logging(LOG_NOTICE, "SMART capable device not found");
	    /* Go on a sabbatical. Exiting makes init unhappy. */
	    /* JUNOS begin */
	    /* Read the config file even though smartd is going on 
	     * sabbatical. Have indication that smartd is going on 
	     * sabbatical.
	     */  
	    readconfigfile();
	    smartd_sleep_infinitely = 1; 
	    /* JUNOS end */
            while (1) {
                sleep (0xFFFFFFFF);
            }
	}
    }

    checkdevices (atadevicesptr); 
    logging (LOG_ERR, "%s: Unexpected return from checkdevices", __FUNCTION__);
    return 1;
}
